@@ -1,6 +1,7 @@
-use crate::Token;
-use std::convert::{Infallible, TryFrom, TryInto};
-use std::iter;
+use crate::{
+    Deserialize, DeserializeError, Deserializer, Serialize, Serializer, Token, ValidationError,
+};
+use std::convert::{TryFrom, TryInto};
 use thiserror::Error;
 
 /// Integer ranging from `-(2^63)` to `(2^64)-1`.
@@ -272,6 +273,12 @@ impl Bin {
     }
 }
 
+#[derive(Debug, Clone, PartialEq, Eq)]
+pub struct Ext {
+    pub r#type: i8,
+    pub data: Vec<u8>,
+}
+
 #[doc(hidden)]
 #[derive(Clone, PartialEq, Debug)]
 pub enum Value {
@@ -284,7 +291,7 @@ pub enum Value {
     Bin(Bin),
     Array(Vec<Value>),
     Map(Vec<(Value, Value)>),
-    Ext(i8, Vec<u8>),
+    Ext(Ext),
 }
 
 impl From<bool> for Value {
@@ -395,6 +402,12 @@ impl From<Bin> for Value {
     }
 }
 
+impl From<Ext> for Value {
+    fn from(v: Ext) -> Self {
+        Self::Ext(v)
+    }
+}
+
 impl From<Vec<Value>> for Value {
     fn from(v: Vec<Value>) -> Self {
         Self::Array(v)
@@ -513,7 +526,7 @@ impl Value {
         matches!(self, Self::Map(_))
     }
     pub fn is_ext(&self) -> bool {
-        matches!(self, Self::Ext(_, _))
+        matches!(self, Self::Ext(_))
     }
     pub fn as_bool(&self) -> Option<bool> {
         match self {
@@ -587,143 +600,22 @@ impl Value {
             _ => None,
         }
     }
-    pub fn as_ext(&self) -> Option<(i8, &[u8])> {
+    pub fn as_ext(&self) -> Option<&Ext> {
         match self {
-            Self::Ext(tag, v) => Some((*tag, v)),
+            Self::Ext(v) => Some(v),
             _ => None,
         }
     }
-    pub fn as_ext_mut(&mut self) -> Option<(i8, &mut Vec<u8>)> {
+    pub fn as_ext_mut(&mut self) -> Option<&mut Ext> {
         match self {
-            Self::Ext(tag, v) => Some((*tag, v)),
+            Self::Ext(v) => Some(v),
             _ => None,
         }
     }
 }
 
-struct Serializer {
-    stack: Vec<(usize, Value)>,
-}
-
-#[derive(Error, Debug, Clone)]
-#[error("any of (hand-written) implementations of Serialize should be incorrect")]
-struct SerializerError;
-
-impl Serializer {
-    pub fn new() -> Self {
-        Self { stack: vec![] }
-    }
-
-    pub fn finish(mut self) -> Value {
-        assert_eq!(self.stack.len(), 1, "{}", SerializerError);
-        let (count, value) = self.stack.pop().unwrap();
-        assert_eq!(count, 0, "{}", SerializerError);
-        value
-    }
-
-    fn push_atom(&mut self, mut v: Value) {
-        loop {
-            if let Some((count, container)) = self.stack.last_mut() {
-                if let Some(array) = container.as_array_mut() {
-                    assert!(*count < array.len(), "{}", SerializerError);
-                    array[*count] = v;
-                    *count += 1;
-                    if *count == array.len() {
-                        v = self.stack.pop().unwrap().1;
-                        continue;
-                    }
-                } else if let Some(map) = container.as_map_mut() {
-                    assert!(*count < map.len() * 2, "{}", SerializerError);
-                    if *count % 2 == 0 {
-                        map[*count / 2].0 = v;
-                    } else {
-                        map[*count / 2].1 = v;
-                    }
-                    *count += 1;
-                    if *count == map.len() * 2 {
-                        v = self.stack.pop().unwrap().1;
-                        continue;
-                    }
-                } else {
-                    panic!("{}", SerializerError);
-                }
-            } else {
-                self.stack.push((0, v));
-            }
-            break;
-        }
-    }
-}
-
-impl crate::Serializer for Serializer {
-    type Error = std::convert::Infallible;
-    fn serialize_nil(&mut self) -> Result<(), Self::Error> {
-        self.push_atom(Value::Nil);
-        Ok(())
-    }
-    fn serialize_bool(&mut self, v: bool) -> Result<(), Self::Error> {
-        self.push_atom(Value::from(v));
-        Ok(())
-    }
-    fn serialize_int(&mut self, v: Int) -> Result<(), Self::Error> {
-        self.push_atom(Value::Int(v));
-        Ok(())
-    }
-    fn serialize_f32(&mut self, v: f32) -> Result<(), Self::Error> {
-        self.push_atom(Value::from(v));
-        Ok(())
-    }
-    fn serialize_f64(&mut self, v: f64) -> Result<(), Self::Error> {
-        self.push_atom(Value::from(v));
-        Ok(())
-    }
-    fn serialize_str(&mut self, v: &[u8]) -> Result<(), Self::Error> {
-        self.push_atom(Value::Str(Str(v.to_owned())));
-        Ok(())
-    }
-    fn serialize_bin(&mut self, v: &[u8]) -> Result<(), Self::Error> {
-        self.push_atom(Value::Bin(Bin(v.to_owned())));
-        Ok(())
-    }
-    fn serialize_array(&mut self, len: u32) -> Result<(), Self::Error> {
-        if len == 0 {
-            self.push_atom(Value::Array(vec![]));
-        } else {
-            let vec = iter::repeat(Value::Nil).take(len as usize).collect();
-            self.stack.push((0, Value::Array(vec)));
-        }
-        Ok(())
-    }
-    fn serialize_map(&mut self, len: u32) -> Result<(), Self::Error> {
-        if len == 0 {
-            self.push_atom(Value::Map(vec![]));
-        } else {
-            let vec = iter::repeat((Value::Nil, Value::Nil))
-                .take(len as usize)
-                .collect();
-            self.stack.push((0, Value::Map(vec)));
-        }
-        Ok(())
-    }
-    fn serialize_ext(&mut self, tag: i8, v: &[u8]) -> Result<(), Self::Error> {
-        self.push_atom(Value::Ext(tag, v.to_owned()));
-        Ok(())
-    }
-}
-
-#[doc(hidden)]
-pub fn serialize<S: crate::Serialize>(x: &S) -> Value {
-    let mut serializer = Serializer::new();
-    x.serialize(&mut serializer)
-        .unwrap_or_else(|infallible| match infallible {});
-    serializer.finish()
-}
-
-impl crate::Serialize for Value {
-    fn serialize<S>(&self, serializer: &mut S) -> Result<(), S::Error>
-    where
-        S: crate::Serializer,
-    {
+impl Serialize for Value {
+    fn serialize(&self, serializer: &mut Serializer) {
         match self {
             Value::Nil => serializer.serialize_nil(),
             Value::Bool(v) => serializer.serialize_bool(*v),
@@ -733,88 +625,26 @@ impl crate::Serialize for Value {
             Value::Str(v) => serializer.serialize_str(&v.0),
             Value::Bin(v) => serializer.serialize_bin(&v.0),
             Value::Array(v) => {
-                serializer.serialize_array(v.len() as u32)?;
+                serializer.serialize_array(v.len() as u32);
                 for x in v {
-                    x.serialize(serializer)?;
+                    serializer.serialize(x);
                 }
-                Ok(())
             }
             Value::Map(v) => {
-                serializer.serialize_map(v.len() as u32)?;
+                serializer.serialize_map(v.len() as u32);
                 for (k, v) in v {
-                    k.serialize(serializer)?;
-                    v.serialize(serializer)?;
+                    serializer.serialize(k);
+                    serializer.serialize(v);
                 }
-                Ok(())
             }
-            Value::Ext(tag, data) => serializer.serialize_ext(*tag, data),
+            Value::Ext(v) => serializer.serialize_ext(v.r#type, &v.data),
         }
     }
 }
 
-impl Value {
-    fn into_iter(self) -> Box<dyn Iterator<Item = Token>> {
-        match self {
-            Value::Nil => Box::new(iter::once(Token::Nil)),
-            Value::Bool(v) => Box::new(iter::once(Token::Bool(v))),
-            Value::Int(v) => Box::new(iter::once(Token::Int(v))),
-            Value::F32(v) => Box::new(iter::once(Token::F32(v))),
-            Value::F64(v) => Box::new(iter::once(Token::F64(v))),
-            Value::Str(v) => Box::new(iter::once(Token::Str(v))),
-            Value::Bin(v) => Box::new(iter::once(Token::Bin(v))),
-            Value::Array(v) => Box::new(
-                iter::once(Token::Array(v.len() as u32))
-                    .chain(v.into_iter().flat_map(Value::into_iter)),
-            ),
-            Value::Map(v) => Box::new(
-                iter::once(Token::Map(v.len() as u32)).chain(
-                    v.into_iter()
-                        .flat_map(|(k, v)| k.into_iter().chain(v.into_iter())),
-                ),
-            ),
-            Value::Ext(tag, data) => Box::new(iter::once(Token::Ext(tag, data))),
-        }
-    }
-}
-
-struct Deserializer {
-    iter: Box<dyn Iterator<Item = Token>>,
-}
-
-impl Deserializer {
-    pub fn new(value: Value) -> Self {
-        Self {
-            iter: value.into_iter(),
-        }
-    }
-}
-
-impl crate::Deserializer for Deserializer {
-    type Error = std::convert::Infallible;
-
-    fn deserialize(&mut self) -> Result<Token, Self::Error> {
-        let token = self
-            .iter
-            .next()
-            .expect("any of (hand-written) implementations of Deserialize should be incorrect");
-        Ok(token)
-    }
-}
-
-#[doc(hidden)]
-pub fn deserialize<D: crate::Deserialize>(
-    value: Value,
-) -> Result<D, crate::DeserializeError<Infallible>> {
-    let mut deserializer = Deserializer::new(value);
-    D::deserialize(&mut deserializer)
-}
-
-impl crate::Deserialize for Value {
-    fn deserialize<D>(deserializer: &mut D) -> Result<Self, crate::DeserializeError<D::Error>>
-    where
-        D: crate::Deserializer,
-    {
-        let x = match deserializer.deserialize()? {
+impl Deserialize for Value {
+    fn deserialize(deserializer: &mut Deserializer) -> Result<Self, DeserializeError> {
+        let x = match deserializer.deserialize_token()? {
             Token::Nil => Value::Nil,
             Token::Bool(v) => v.into(),
             Token::Int(v) => v.into(),
@@ -823,28 +653,37 @@ impl crate::Deserialize for Value {
             Token::Str(v) => v.into(),
             Token::Bin(v) => v.into(),
             Token::Array(len) => {
-                let mut vec = vec![];
+                let mut vec: Vec<Value> = vec![];
                 for _ in 0..len {
-                    vec.push(Value::deserialize(deserializer)?);
+                    vec.push(deserializer.deserialize()?);
                 }
                 vec.into()
             }
             Token::Map(len) => {
-                let mut map = vec![];
+                let mut map: Vec<(Value, Value)> = vec![];
                 for _ in 0..len {
-                    map.push((
-                        Value::deserialize(deserializer)?,
-                        Value::deserialize(deserializer)?,
-                    ));
+                    map.push((deserializer.deserialize()?, deserializer.deserialize()?));
                 }
                 map.into()
             }
-            Token::Ext(tag, data) => Value::Ext(tag, data),
+            Token::Ext(v) => v.into(),
         };
         Ok(x)
     }
 }
 
+#[doc(hidden)]
+pub fn serialize<S: Serialize>(x: &S) -> Value {
+    let buf = crate::serialize(x);
+    crate::deserialize(&buf).unwrap()
+}
+
+#[doc(hidden)]
+pub fn deserialize<D: Deserialize>(value: Value) -> Result<D, DeserializeError> {
+    let buf = crate::serialize(value);
+    crate::deserialize::<D>(&buf)
+}
+
 /// A special type for serializing and deserializing the `nil` object.
 ///
 /// In our data model `()` does not represent the `nil` object because `()` should be zero-byte but `nil` has a size.
@@ -852,23 +691,17 @@ impl crate::Deserialize for Value {
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
 pub struct Nil;
 
-impl crate::Serialize for Nil {
-    fn serialize<S>(&self, serializer: &mut S) -> Result<(), S::Error>
-    where
-        S: crate::Serializer,
-    {
+impl Serialize for Nil {
+    fn serialize(&self, serializer: &mut Serializer) {
         serializer.serialize_nil()
     }
 }
 
-impl crate::Deserialize for Nil {
-    fn deserialize<D>(deserializer: &mut D) -> Result<Self, crate::DeserializeError<D::Error>>
-    where
-        D: crate::Deserializer,
-    {
-        let token = deserializer.deserialize()?;
+impl Deserialize for Nil {
+    fn deserialize(deserializer: &mut Deserializer) -> Result<Self, DeserializeError> {
+        let token = deserializer.deserialize_token()?;
         if token != Token::Nil {
-            return Err(crate::DeserializeError::InvalidType);
+            return Err(ValidationError.into());
         }
         Ok(Self)
     }
@@ -878,15 +711,12 @@ impl crate::Deserialize for Nil {
 #[derive(Debug, Clone, Copy, PartialEq, Eq)]
 pub struct Any;
 
-impl crate::Deserialize for Any {
-    fn deserialize<D>(deserializer: &mut D) -> Result<Self, crate::DeserializeError<D::Error>>
-    where
-        D: crate::Deserializer,
-    {
+impl Deserialize for Any {
+    fn deserialize(deserializer: &mut Deserializer) -> Result<Self, DeserializeError> {
         let mut count = 1;
         while count > 0 {
             count -= 1;
-            match deserializer.deserialize()? {
+            match deserializer.deserialize_token()? {
                 Token::Nil
                 | Token::Bool(_)
                 | Token::Int(_)
@@ -894,7 +724,7 @@ impl crate::Deserialize for Any {
                 | Token::F64(_)
                 | Token::Str(_)
                 | Token::Bin(_)
-                | Token::Ext(_, _) => {}
+                | Token::Ext(_) => {}
                 Token::Array(len) => {
                     count += len;
                 }
